@@ -16,17 +16,19 @@ import os
 import shutil
 import sys
 import time
+import platform
 from concurrent.futures import Future, ThreadPoolExecutor
 from contextlib import contextmanager
 from datetime import datetime
 from enum import Enum
-from typing import Any, ContextManager, IO, List, Mapping, NamedTuple, Optional, Tuple
+from typing import Any, ContextManager, IO, List, Mapping, NamedTuple, Optional, Tuple, Generator
 
 import dateutil.parser
 import progressbar
 
 from harmony.auth import create_session, validate_auth
 from harmony.config import Config, Environment
+from harmony import __version__ as harmony_version
 
 progressbar_widgets = [
     ' [ Processing: ', progressbar.Percentage(), ' ] ',
@@ -361,6 +363,51 @@ class Client:
 
         return params
 
+    def _headers(self) -> dict:
+        """
+        Create (if needed) and return a dictionary of headers.
+        Code partially adaped from:
+            https://github.com/requests/toolbelt/blob/master/requests_toolbelt/utils/user_agent.py
+        """
+        if 'headers' not in self.__dict__:
+            session = self._session()
+            existing_user_agent_header = session.headers.get('User-Agent')
+            if existing_user_agent_header:
+                user_agent_content = set([existing_user_agent_header])
+            else:
+                user_agent_content = set([])
+
+            # Get harmony package info
+            user_agent_content.add(f'harmony-py/{harmony_version}')
+
+            # Get platform info
+            try:
+                p_system = platform.system()
+                p_release = platform.release()
+                user_agent_content.add(f'{p_system}/{p_release}')
+            except Exception as e:
+                print("Following exception was caught "
+                      "when building user-agent headers for harmony-py:")
+                print(e)
+
+            # Get implementation info
+            try:
+                implementation = platform.python_implementation()
+                implementation_version = platform.python_version()
+                user_agent_content.add(f'{implementation}/{implementation_version}')
+            except Exception as e:
+                print("Following exception was caught "
+                      "when building user-agent headers for harmony-py:")
+                print(e)
+
+            # Build headers
+            if user_agent_content:
+                self.headers = {'User-Agent': ' '.join(user_agent_content)}
+            else:
+                self.headers = {}
+
+        return self.headers
+
     def _spatial_subset_params(self, request: Request) -> list:
         """Creates a dictionary of spatial subset query parameters."""
         if request.spatial:
@@ -442,6 +489,7 @@ class Client:
         job_id = None
         session = self._session()
         params = self._params(request)
+        headers = self._headers()
 
         with self._files(request) as files:
             if files:
@@ -455,11 +503,11 @@ class Client:
                 file_items = [(k, v) for k, v in files.items()]
                 response = session.post(
                     self._submit_url(request),
-                    files=param_items + file_items)
+                    files=param_items + file_items,
+                    headers=headers)
             else:
-                response = session.get(self._submit_url(request), params=params)
+                response = session.get(self._submit_url(request), params=params, headers=headers)
         if response.ok:
-            print(response.json())
             job_id = (response.json())['jobID']
         else:
             response.raise_for_status()
@@ -585,6 +633,7 @@ class Client:
         Args:
             job_id: UUID string for the job you wish to interrogate.
             show_progress: Whether a progress bar should show via stdout.
+            link_type: The type of link to output, s3:// or https://
 
         Returns:
             The job's complete json output.
@@ -593,13 +642,46 @@ class Client:
             self.wait_for_processing(job_id, show_progress)
         except ProcessingFailedException:
             pass
-        response = self._session().get(self._status_url(job_id))
+        response = self._session().get(self._status_url(job_id, link_type))
         return response.json()
 
+    def _get_json(self, url: str) -> str:
+        """Gets and parses the JSON at the given URL
+
+        Args:
+            url: The URL to get
+
+        Returns:
+            The parsed JSON contents of the given URL
+        """
+        return self._session().get(url).json()
+
+    def _result_pages(self,
+                      job_id: str,
+                      show_progress: bool = False,
+                      link_type: LinkType = LinkType.https) -> Generator[object, None, None]:
+        """Yields each page of results for the provided job ID
+
+        Args:
+            job_id: UUID string for the job to be fetched
+            show_progress: Whether a progress bar should show via stdout.
+            link_type: The type of link to output, s3:// or https://
+
+        Returns:
+            A generator for each page of results, loaded on demand
+        """
+        self.wait_for_processing(job_id, show_progress)
+        next_url = self._status_url(job_id, link_type)
+        while next_url:
+            response = self._get_json(next_url)
+            yield response
+            links = response.get('links', [])
+            next_url = next((x['href'] for x in links if x['rel'] == 'next'), None)
+
     def result_urls(self,
                     job_id: str,
                     show_progress: bool = False,
-                    link_type: LinkType = LinkType.https) -> List:
+                    link_type: LinkType = LinkType.https) -> Generator[str, None, None]:
         """Retrieve the data URLs for a job.
 
         The URLs include links to all of the jobs data output. Blocks until the Harmony job is
@@ -608,13 +690,15 @@ class Client:
         Args:
             job_id: UUID string for the job you wish to interrogate.
             show_progress: Whether a progress bar should show via stdout.
+            link_type: The type of link to output, s3:// or https://
 
         Returns:
             The job's complete list of data URLs.
         """
-        data = self.result_json(job_id, show_progress, link_type)
-        urls = [x['href'] for x in data.get('links', []) if x['rel'] == 'data']
-        return urls
+        for page in self._result_pages(job_id, show_progress, link_type):
+            for link in page.get('links', []):
+                if link['rel'] == 'data':
+                    yield link['href']
 
     def _download_file(self, url: str, directory: str = '', overwrite: bool = False) -> str:
         chunksize = int(self.config.DOWNLOAD_CHUNK_SIZE)
@@ -657,7 +741,7 @@ class Client:
     def download_all(self,
                      job_id: str,
                      directory: str = '',
-                     overwrite: bool = False) -> List[Future]:
+                     overwrite: bool = False) -> Generator[Future, None, None]:
         """Using a job_id, fetches all the data files from a finished job.
 
         After this method is able to contact Harmony and query a finished job, it will
@@ -687,10 +771,8 @@ class Client:
             A list of Futures, each of which will return the filename (with path) for each
             result.
         """
-        urls = self.result_urls(job_id, show_progress=False) or []
-        return [
-            self.executor.submit(self._download_file, url, directory, overwrite) for url in urls
-        ]
+        for url in self.result_urls(job_id, show_progress=False) or []:
+            yield self.executor.submit(self._download_file, url, directory, overwrite)
 
     def stac_catalog_url(self,
                          job_id: str,
@@ -703,6 +785,7 @@ class Client:
         Args:
             job_id: UUID string for the job you wish to interrogate.
             show_progress: Whether a progress bar should show via stdout.
+            link_type: The type of link to output, s3:// or https://
 
         Returns:
             A STAC catalog URL.
